@@ -3,13 +3,28 @@
 graph a MAVLink log file
 Andrew Tridgell August 2011
 '''
+from __future__ import print_function
+from builtins import input
+from builtins import range
 
-import sys, struct, time, os, datetime
-import math, re
+import datetime
 import matplotlib
+import os
+import re
+import sys
+import time
 from math import *
 
-from pymavlink.mavextra import *
+try:
+    from pymavlink.mavextra import *
+except:
+    print("WARNING: Numpy missing, mathematical notation will not be supported.")
+
+# cope with rename of raw_input in python3
+try:
+    input = raw_input
+except NameError:
+    pass
 
 colourmap = {
     'apm' : {
@@ -85,7 +100,7 @@ def plotit(x, y, fields, colors=[]):
             color = 'red'
         (tz, tzdst) = time.tzname
         if axes[i] == 2:
-            if ax2 == None:
+            if ax2 is None:
                 ax2 = ax1.twinx()
             ax = ax2
             if not args.xaxis:
@@ -134,6 +149,7 @@ def plotit(x, y, fields, colors=[]):
     if empty:
         print("No data to graph")
         return
+    return fig
 
 
 from argparse import ArgumentParser
@@ -154,6 +170,7 @@ parser.add_argument("--flightmode", default=None,
                     help="Choose the plot background according to the active flight mode of the specified type, e.g. --flightmode=apm for ArduPilot or --flightmode=px4 for PX4 stack logs.  Cannot be specified with --xaxis.")
 parser.add_argument("--dialect", default="ardupilotmega", help="MAVLink dialect")
 parser.add_argument("--output", default=None, help="provide an output format")
+parser.add_argument("--timeshift", type=float, default=0, help="shift time on first graph in seconds")
 parser.add_argument("logs_fields", metavar="<LOG or FIELD>", nargs="+")
 args = parser.parse_args()
 
@@ -231,7 +248,7 @@ def add_data(t, msg, vars, flightmode):
         y[i].append(v)
         x[i].append(xv)
 
-def process_file(filename):
+def process_file(filename, timeshift):
     '''process one file'''
     print("Processing %s" % filename)
     mlog = mavutil.mavlink_connection(filename, notimestamps=args.notimestamps, zero_time_base=args.zero_time_base, dialect=args.dialect)
@@ -240,7 +257,12 @@ def process_file(filename):
     while True:
         msg = mlog.recv_match(args.condition)
         if msg is None: break
-        tdays = matplotlib.dates.date2num(datetime.datetime.fromtimestamp(msg._timestamp))
+        try:
+            tdays = matplotlib.dates.date2num(datetime.datetime.fromtimestamp(msg._timestamp+timeshift))
+        except ValueError:
+            # this can happen if the log is corrupt
+            # ValueError: year is out of range
+            break
         add_data(tdays, msg, mlog.messages, mlog.flightmode)
 
 if len(filenames) == 0:
@@ -256,9 +278,12 @@ if args.labels is not None:
 else:
     labels = None
 
+timeshift = args.timeshift
+
 for fi in range(0, len(filenames)):
     f = filenames[fi]
-    process_file(f)
+    process_file(f, timeshift)
+    timeshift = 0
     for i in range(0, len(x)):
         if first_only[i] and fi != 0:
             x[i] = []
@@ -271,14 +296,22 @@ for fi in range(0, len(filenames)):
         col = colors[:]
     else:
         col = colors[fi*len(fields):]
-    plotit(x, y, lab, colors=col)
+    fig = plotit(x, y, lab, colors=col)
     for i in range(0, len(x)):
         x[i] = []
         y[i] = []
 if args.output is None:
     pylab.show()
     pylab.draw()
-    raw_input('press enter to exit....')
+    input('press enter to exit....')
 else:
-    pylab.legend(loc=2,prop={'size':8})
-    pylab.savefig(args.output, bbox_inches='tight', dpi=200)
+    fname, fext = os.path.splitext(args.output)
+    if fext == '.html':
+        import mpld3
+        html = mpld3.fig_to_html(fig)
+        f_out = open(args.output, 'w')
+        f_out.write(html)
+        f_out.close()
+    else:
+        pylab.legend(loc=2,prop={'size':8})
+        pylab.savefig(args.output, bbox_inches='tight', dpi=200)
