@@ -3,21 +3,20 @@
 '''
 extract mavlink parameter values
 '''
+from __future__ import print_function
 
-import sys, time, os
 
-from optparse import OptionParser
-parser = OptionParser("mavparms.py [options]")
-parser.add_option("-c", "--changes", dest="changesOnly", default=False, action="store_true", help="Show only changes to parameters.")
+import time
 
-(opts, args) = parser.parse_args()
+from argparse import ArgumentParser
+parser = ArgumentParser(description=__doc__)
+parser.add_argument("-c", "--changes", dest="changesOnly", default=False, action="store_true", help="Show only changes to parameters.")
+parser.add_argument("logs", metavar="LOG", nargs="+")
+
+args = parser.parse_args()
 
 from pymavlink import mavutil
 
-if len(args) < 1:
-    print("Usage: mavparms.py [options] <LOGFILE...>")
-    sys.exit(1)
-
 parms = {}
 
 def mavparms(logfile):
@@ -38,17 +37,17 @@ def mavparms(logfile):
             pname = m.Name
             value = m.Value
         if len(pname) > 0:
-            if opts.changesOnly is True and pname in parms and parms[pname] != value:
+            if args.changesOnly is True and pname in parms and parms[pname] != value:
                 print("%s %-15s %.6f -> %.6f" % (time.asctime(time.localtime(m._timestamp)), pname, parms[pname], value))
-            
+
             parms[pname] = value
 
 total = 0.0
-for filename in args:
+for filename in args.logs:
     mavparms(filename)
 
-if (opts.changesOnly is False):
-    keys = parms.keys()
+if (args.changesOnly is False):
+    keys = list(parms.keys())
     keys.sort()
     for p in keys:
         print("%-15s %.6f" % (p, parms[p]))
